from typing import Optional
import click
from pathlib import Path
from . import commands
from importlib import metadata


DEFAULT_CONFIG_PATH = Path(".daft.toml")


def get_config_path(config: Optional[Path]) -> Path:
    if config:
        if not config.exists():
            raise click.UsageError("Config file does not exist.")
    else:
        config = DEFAULT_CONFIG_PATH
        if not config.exists():
            raise click.UsageError(
                f"No default '{DEFAULT_CONFIG_PATH}' file found in current directory."
            )
    return config


def assert_identity_file_path(identity_file: Optional[Path]):
    if not identity_file:
        return
    if not identity_file.exists():
        raise click.UsageError("Identity file does not exist.")


def assert_working_dir(working_dir: Path):
    if not working_dir.exists():
        raise click.UsageError("Working dir does not exist.")
    if not working_dir.is_dir():
        raise click.UsageError("Working dir must be a directory.")


def get_new_configuration_file_path(name: Optional[Path]) -> Path:
    name = name or DEFAULT_CONFIG_PATH
    if name.exists():
        raise click.UsageError(f"A configuration file at path {name} already exists.")
    return name


def identity_file_option(func):
    return click.option(
        "--identity-file",
        "-i",
        required=False,
        type=Path,
        help="Path to the identity file.",
    )(func)


def working_dir_option(func):
    return click.option(
        "--working-dir",
        "-w",
        required=True,
        type=Path,
        help="Path to the working directory.",
    )(func)


def init_config_file_name_argument(func):
    return click.argument(
        "name",
        required=False,
        type=Path,
    )(func)


def config_option(func):
    return click.option(
        "--config",
        "-c",
        required=False,
        type=Path,
        help=f"Path to the configuration file; defaults to {DEFAULT_CONFIG_PATH}.",
    )(func)


def cmd_args_argument(func):
    return click.argument("cmd_args", nargs=-1, type=click.UNPROCESSED, required=True)(func)


def init_config_command(func):
    return click.command("init-config", help="Create a new configuration file.")(func)


def up_command(func):
    return click.command("up", help="Spin the cluster up.")(func)


def down_command(func):
    return click.command("down", help="Spin the cluster down.")(func)


def list_command(func):
    return click.command("list", help="List all running clusters.")(func)


def dashboard_command(func):
    return click.command(
        "dashboard",
        help="Enable port-forwarding between a cluster and your local machine.",
    )(func)


def submit_command(func):
    return click.command("submit", help="Submit a job to the specified cluster.")(func)


@init_config_command
@init_config_file_name_argument
def init_config(name: Optional[Path]):
    name = get_new_configuration_file_path(name)
    commands.init_config(name)


@up_command
@config_option
def up(config: Optional[Path]):
    config = get_config_path(config)
    commands.up(config)


@list_command
def list():
    commands.list()


@dashboard_command
@config_option
@identity_file_option
def dashboard(
    config: Optional[Path],
    identity_file: Optional[Path],
):
    config = get_config_path(config)
    assert_identity_file_path(identity_file)
    commands.dashboard(config, identity_file)


@submit_command
@config_option
@working_dir_option
@identity_file_option
@cmd_args_argument
def submit(
    config: Optional[Path],
    identity_file: Optional[Path],
    working_dir: Path,
    cmd_args: tuple[str],
):
    config = get_config_path(config)
    assert_identity_file_path(identity_file)
    assert_working_dir(working_dir)
    commands.submit(config, identity_file, working_dir, cmd_args)


@down_command
@config_option
def down(config: Optional[Path]):
    config = get_config_path(config)
    commands.down(config)


@click.group(help=metadata.metadata("daft-launcher").get("Summary"))
@click.version_option(version=metadata.version("daft-launcher"))
def cli(): ...


def run_cli():
    cli.add_command(init_config)
    cli.add_command(up)
    cli.add_command(list)
    cli.add_command(dashboard)
    cli.add_command(submit)
    cli.add_command(down)
    cli()



from typing import List, Optional, Any
from pathlib import Path
import subprocess
from . import configs, helpers
from ray.autoscaler import sdk as ray_sdk
from ray import job_submission
import click
import time


AWS_TEMPLATE_PATH = Path(__file__).parent / "templates" / "aws.toml"


def init_config(name: Path):
    with open(AWS_TEMPLATE_PATH) as template_f:
        with open(name, "w") as config_f:
            config_f.write(template_f.read())


def up(config: Path):
    final_config = configs.get_merged_config(config)
    ray_sdk.create_or_update_cluster(
        final_config,
        no_restart=False,
        restart_only=False,
        no_config_cache=True,
    )
    print(f"Head node IP: {ray_sdk.get_head_node_ip(final_config)}")
    print("Successfully spun the cluster up.")


def list():
    instance_groups: List[List[Any]] = helpers.run_aws_command(
        [
            "aws",
            "ec2",
            "describe-instances",
            "--region",
            "us-west-2",
            "--filters",
            "Name=tag:ray-node-type,Values=*",
            "--query",
            "Reservations[*].Instances[*].{Instance:InstanceId,State:State.Name,Tags:Tags,Ip:PublicIpAddress}",
        ]
    )
    state_to_name_map: dict[str, List[tuple]] = {}
    for instance_group in instance_groups:
        for instance in instance_group:
            state = instance["State"]
            instance_id: str = instance["Instance"]
            ip: str = instance["Ip"]
            name = None
            node_type = None
            for tag in instance["Tags"]:
                if tag["Key"] == "ray-node-type":
                    node_type = tag["Value"]
                if tag["Key"] == "ray-cluster-name":
                    name = tag["Value"]
            assert name is not None
            assert node_type is not None
            if state in state_to_name_map:
                state_to_name_map[state].append((name, instance_id, node_type, ip))
            else:
                state_to_name_map[state] = [(name, instance_id, node_type, ip)]
    for state, data in state_to_name_map.items():
        print(f"{state.capitalize()}:")
        for name, instance_id, node_type, ip in data:
            formatted_name = f""
            print(
                f"\t - {name}, {node_type}, {instance_id}" + (f", {ip}" if ip else "")
            )


def dashboard(
    config: Path,
    identity_file: Optional[Path],
):
    subprocess.run(
        helpers.ssh_command(
            helpers.get_ip(config), Path(identity_file) if identity_file else None
        ),
        close_fds=True,
        capture_output=False,
        text=False,
    )


def submit(
    config: Path,
    identity_file: Optional[Path],
    working_dir: Path,
    cmd_args: tuple[str],
):
    cmd = " ".join([arg for arg in cmd_args])
    process = subprocess.Popen(
        helpers.ssh_command(
            helpers.get_ip(config), Path(identity_file) if identity_file else None
        ),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    try:
        working_dir_path = Path(working_dir).absolute()
        client = None
        tries = 0
        max_tries = 3
        while tries <= max_tries:
            try:
                client = job_submission.JobSubmissionClient("http://localhost:8265")
                break
            except Exception as e:
                tries += 1
                if tries >= max_tries:
                    raise e
                time.sleep(1)
        assert client
        id = client.submit_job(
            entrypoint=cmd,
            runtime_env={
                "working_dir": working_dir_path.absolute(),
            }
            if working_dir
            else None,
        )
        print(f"Job ID: {id}")
    finally:
        process.terminate()


def down(config: Path):
    final_config = configs.get_merged_config(config)
    ray_sdk.teardown_cluster(final_config)
    print("Successfully spun the cluster down.")



from pathlib import Path
import tomllib
from typing import Optional
import click
import yaml


RAY_DEFAULT_CONFIGS_PATH = Path(__file__).parent / "ray_default_configs"


def get_ray_config(
    provider: str,
) -> tuple[dict, list[tuple[str, list[list[str]], bool]]]:
    if provider == "aws":
        ray_config_path = RAY_DEFAULT_CONFIGS_PATH / "aws.yaml"
        setup_mappings = [
            ("name", [["cluster_name"]], False),
            ("provider", [["provider", "type"]], True),
            ("region", [["provider", "region"]], False),
            ("ssh_user", [["auth", "ssh_user"]], False),
            (
                "number_of_workers",
                [
                    ["max_workers"],
                    ["available_node_types", "ray.worker.default", "min_workers"],
                    ["available_node_types", "ray.worker.default", "max_workers"],
                ],
                False,
            ),
            (
                "instance_type",
                [
                    [
                        "available_node_types",
                        "ray.head.default",
                        "node_config",
                        "InstanceType",
                    ],
                    [
                        "available_node_types",
                        "ray.worker.default",
                        "node_config",
                        "InstanceType",
                    ],
                ],
                False,
            ),
            (
                "image_id",
                [
                    [
                        "available_node_types",
                        "ray.head.default",
                        "node_config",
                        "ImageId",
                    ],
                    [
                        "available_node_types",
                        "ray.worker.default",
                        "node_config",
                        "ImageId",
                    ],
                ],
                False,
            ),
        ]
    else:
        raise click.UsageError(f"Cloud provider {provider} not found")
    with open(ray_config_path, "rb") as stream:
        return yaml.safe_load(stream), setup_mappings


def get_custom_config(config_path: Path) -> tuple[dict, str]:
    with open(config_path, "rb") as stream:
        custom_config = tomllib.load(stream)
        if "setup" not in custom_config:
            raise click.UsageError("No setup section found in config file.")
            if "provider" not in custom_config["setup"]:
                raise click.UsageError(
                    "Please provide a cloud provider in the config file."
                )
        provider = custom_config["setup"]["provider"]
        return custom_config, provider


def merge(
    custom_config: dict,
    ray_config: dict,
    setup_mappings: list,
) -> dict:
    if "setup" not in custom_config:
        raise click.UsageError("No setup section found in config file.")
    setup: dict = custom_config["setup"]
    for a, b, required in setup_mappings:
        value = setup.get(a)
        if not value and required:
            raise click.UsageError(
                f"Required field {a} not found in custom config but is required."
            )
        elif not value:
            continue
        for b_i in b:
            y = ray_config
            for b_ii in b_i[:-1]:
                y = y[b_ii]
            y[b_i[-1]] = value

    setup_commands: list[str] = ray_config["setup_commands"]
    if "dependencies" in setup:
        quoted_dependencies = [f'"{dep}"' for dep in setup["dependencies"]]
        dependencies = " ".join(quoted_dependencies)
        uv_install_command = f"uv pip install {dependencies}"
        setup_commands.append(uv_install_command)
    if "run" in custom_config:
        if "setup_commands" in custom_config["run"]:
            setup_commands.extend(custom_config["run"]["setup_commands"])
    return ray_config


def get_merged_config(config_path: Path) -> dict:
    custom_config, provider = get_custom_config(config_path)
    ray_config, setup_mappings = get_ray_config(provider)
    return merge(custom_config, ray_config, setup_mappings)



from typing import List, Optional, Any
from pathlib import Path
import subprocess
import json
from . import configs
import click


def get_ip(config: Path):
    final_config = configs.get_merged_config(config)
    name = final_config["cluster_name"]
    instance_groups: List[List[Any]] = run_aws_command(
        [
            "aws",
            "ec2",
            "describe-instances",
            "--region",
            "us-west-2",
            "--filters",
            "Name=tag:ray-node-type,Values=*",
            "--query",
            "Reservations[*].Instances[*].{State:State.Name,Tags:Tags,Ip:PublicIpAddress}",
        ],
    )
    state_to_ips_mapping = find_ip(instance_groups, name)
    if 'running' not in state_to_ips_mapping:
        raise click.UsageError(
            f"The cluster {name} is not running; cannot connect to it."
        )
    assert len(state_to_ips_mapping['running']) <= 1
    if state_to_ips_mapping['running']:
        ip = state_to_ips_mapping['running'][0]
    else:
        raise click.UsageError(
            f"The cluster {name} is not running; cannot connect to it."
        )
    if not ip:
        raise click.UsageError(
            f"The cluster {name} does not have a public IP address available."
        )
    return ip


def run_aws_command(args: list[str]) -> Any:
    result = subprocess.run(args, capture_output=True, text=True)
    if result.returncode != 0:
        if "Token has expired" in result.stderr:
            raise click.UsageError(
                "AWS token has expired. Please run `aws login`, `aws sso login`, or some other command to refresh it."
            )
    if result.stdout == "":
        raise click.UsageError(
            f"Failed to parse AWS command into json (empty response string)"
        )
    try:
        return json.loads(result.stdout)
    except json.JSONDecodeError as e:
        raise click.UsageError(f"Failed to parse AWS command output: {result.stdout}")


def find_ip(instance_groups: List[List[Any]], name: str) -> dict[str, list[Optional[str]]]:
    ip = None
    state_to_ips_mapping: dict[str, list[Optional[str]]] = {}
    def insert(state: str, ip: Optional[str]):
        if state in state_to_ips_mapping:
            state_to_ips_mapping[state].append(ip)
        else:
            state_to_ips_mapping[state] = [ip]

    for instance_group in instance_groups:
        for instance in instance_group:
            is_head = False
            cluster_name = None
            state = instance["State"]
            for tag in instance["Tags"]:
                if tag["Key"] == "ray-cluster-name":
                    cluster_name = tag["Value"]
                elif tag["Key"] == "ray-node-type":
                    is_head = tag["Value"] == "head"
            if is_head and cluster_name == name:
                insert(state, instance["Ip"])

    if not state_to_ips_mapping:
        raise click.UsageError(f"The IP of the cluster with the name '{name}' not found.")

    return state_to_ips_mapping


def ssh_command(
    ip: str,
    pub_key: Optional[Path] = None,
) -> list[str]:
    return (
        [
            "ssh",
            "-N",
            "-L",
            "8265:localhost:8265",
        ]
        + (["-i", str(pub_key)] if pub_key else [])
        + [
            f"ec2-user@{ip}",
        ]
    )
